#
# mbox.rb
#
#   Copyright (c) 1999,2000 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
#
#   This program is free software.
#   You can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

require 'amstd/must'
require 'amstd/bug'
require 'amstd/futils'
require 'amstd/extmod'

require 'tmail/tmail'


module EST

  class CMailset

    Export = %w(
      fetch [] mbox_names
    )

    def initialize
      @set = {}
      @rc = {}
      @inbox = nil
      @draft = nil
      @trash = nil
    end

    def shutdown
      @set.each do |name, mb|
        if Mailbox === mb then
          mb.finalize
        end
      end
    end


    def add_mbox( name, rc )
      rc.must Hash if $DEBUG
      if @rc[name] then
        if @set[name] then
          raise ArgumentError, "realized mbox '#{name}' is re-registered"
        else
          CORE.warn "mbox '#{name}' registered twice"
        end
      end
# puts "mbox added: name=#{name}"
      @rc[name] = rc
    end

    def delete_mbox( name )
      unless @rc[name] then
        raise ArgumentError,
          "try to delete not exist mbox '#{name}'"
      end

      if mb = @set[name] then
        mb.finalize
        @set.delete name
      end
      @rc.delete name
    end

    def add_mbox_group( name, rc )
      if @group[name] then
        raise ArgumentError, "mbox group '#{name}' registered twice"
      end
      ret = MailboxGroup.new( name, rc )
      @group[ name ] = ret
      ret
    end


    def []( name )
      name.must String if $DEBUG

      unless ret = @set[ name ] then
        if rc = @rc[ name ] then
          lo = mkloader( rc )
          @set[ name ] = ret = make_real_mbox( name, rc, lo )
        else
          ret = nombox( name )
        end
      end

      ret
    end

    def nombox( name )
      raise ArgumentError, "no such mbox: #{name}"
    end


    class << self
      def sysmb_attr( label )
        label = label.id2name
        module_eval %-
          def #{label}=( name )
            name.must String
            if mb = @rc[ name ] then
              @#{label} = name
            else
              raise ArgumentError,
                "not exist mbox '\#{name}' was given for #{label}"
            end
          end
        -
        module_eval %-
          def #{label}
            self[ @#{label} ]
          end
        -
      end
    end

    sysmb_attr :inbox
    sysmb_attr :draft
    sysmb_attr :trash


    def mbox_names
      ret = @rc.collect {|name,rc| rc['viewname'] || name }
      ret.sort!
      ret
    end


    def mkloader( rc )
      case rc[ 'type' ]
      when 'mh'
        lo = ::TMail::MhLoader.new( rc['path'] )
      else
        bug! 'mbox type is not MH'
      end

      return lo
    end

    def make_real_mbox( name, rc, lo )
      Mailbox.new( name, rc, lo )
    end

  end

  class Mailset < CMailset
  end



  class MailboxGroup

    def initialize( name, rc )
      @name = name.dup
      @rc = rc
    end

    def exec
      @base = File.expand_path( @rc['path'] )
      procdir @base
    end

    def procdir( dn )
      d = fn = path = nm = rc = nil

      Dir.open( dn ) do |d|
        d.each do |fn|
          next if /\A\./ === fn

          path = File.join( dn, fn )
          next unless File.directory? path

          rc = @rc.dup
          rc['path'] = path
          nm = path.gsub( @base, @name )
          CORE.mailbox.add_mbox nm, rc

          procdir path
        end
      end
    end
  
  end


=begin impl

class Mailbox interface

  implemented by interface

    finalize

    name
    size
    length
    empty?
    []( idx )
    each {|port| .... }
    each_index {|idx| .... }
    each_mail {|mail| .... }
    index( port )
    rindex( port )
    indexes( *args )
    to_a

    load
    reload
    update

    add( port )
    move( idx, to )
    move_of( from, idx )
    remove( idx )
    new
    new_port
    delete_if {|port| .... }
    move_if {|port| .... }

=end impl

  class Mailbox

    def finalize
      save_cache if @loader.can_cache?
    end

    def do_load( no_cache )
      port = ent = nil

      if not no_cache and entries = load_cache then
        entries.each do |ent|
          port = ent.pop
          port.cache = ent.empty? ? nil : ent
          add port
        end
      else
puts 'parsing'
        unread = (@name == 'inbox')
        @loader.each do |p|
          p.unread = unread
          add p
        end
      end
    end
    private :do_load


    private


    CacheVersion = 2

    def load_cache
      fn = "#{CORE['usrdir']}/cache/#{EST::Interface}/#{@name}"
      entries = nil

      begin
        if File.exist? fn then
p fn
          File.open( fn ) do |f|
            ver = f.gets.downcase.strip
            break unless ver == "est cache file version #{CacheVersion}"
puts "version #{CacheVersion}, ok"

            size = f.gets.strip.to_i
            sign = Marshal.load( f.read size )
            break unless cache_signature == sign
puts 'correct cache'

puts "loading cache: #{@name}"
            entries = Marshal.load( f )
          end
        end
unless entries then
puts "cache miss: #{@name}"
end
      rescue
        CORE.warn 'error in reading cache file (may not be fatal)'
        CORE.warn $!.to_s
        entries = nil
      end

      entries
    end

    def save_cache
puts "saving cache: #{@name}"
      fn = "#{CORE['usrdir']}/cache/#{EST::Interface}/#{@name}"
      i = port = nil

      begin
        FileUtils.mkdir_p File.dirname(fn)
        File.open( fn, 'w' ) do |f|
          f.puts "est cache file version #{CacheVersion}"

          str = Marshal.dump( cache_signature )
          f.puts str.size.to_s
          f.write str

          ent = collect{|port|
            i = port.cache ? port.cache.dup : []
            i.push port
            i
          }
          Marshal.dump( ent, f )
        end
      rescue
        $stderr.puts 'error in writing cache file (may not be fatal)'
        $stderr.puts $!.to_s
        if File.file? fn then
          FileUtils.rm_f fn
        end
      end
    end

    # abstract cache_signature()

  end


  class ReferenceError < StandardError ; end


  class VirtualMailbox < Mailbox

    def finalize
    end

    def load( *args )
    end

    def reload
    end

    def update
    end

    def move_of( from, idx )
      raise ReferenceError, "cannot move mail to virtual mbox"
    end

    def new_port
      raise ReferenceError, "virtual mbox cannot make new mail"
    end

  end



  class Mail < TMail::Mail

    def port=( p )
      unless @parsed then
        body
      end
      @port.mv p
      @port = p
    end

    attr :cache

  end


end   # module EST



# post-defines


module TMail

  class Port

    def initialize
      @refc = 0
      @unread = false
      @cache = nil
    end

    attr :cache,  true
    attr :unread, true

    def ref
      @refc += 1
    end

    def unref
      @refc -= 1
#puts "port#unref, refc=#{@refc}"
      if @refc < 1 then
#puts "delete file"
        rm
      else
#puts "not delete"
      end
    end

  end

  class FilePort

    def _dump( limit )
      @filename
    end

    def FilePort._load( str )
      new( str )
    end

  end

  class Loader

    def can_cache?
      false
    end

  end

  class MhLoader

    def can_cache?
      true
    end

  end

end
